#!/usr/bin/env perl

#	Brablenec
#	Copyright (C) 2007 Michal Vaner (vorner@ucw.cz)
#
#	This program is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	he Free Software Foundation; either version 2 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

use warnings;
use strict;

$|=1;

my %settings;
#get the dimensions
my( $pixwidth, $pixheight ) = ( <> =~ /^I:dimensions\s+(\d+)\s+(\d+)/ ) or die "F:no dimensions specified\n";
#get the config file
my( $config ) = ( <> =~ /^I:config\s+"([^"]+)"/ ) or die "F:no config file\n";

{#Config
	open( CONFIG, $config ) or die "F:failed to open config\n";
	print "D:opened config, skipping sections up to brablenec section\n";
	my $ln;
	while( defined( $ln = <CONFIG> ) && ( $ln !~ /^\s*\[brablenec\]\s*$/ ) ) {};

	while( defined( $ln = <CONFIG> ) && ( $ln !~ /^\s*\[.+\]\s*$/ ) ) {
		next if( $ln =~ /^\s*(|#.*)$/ );
		my ( $name, $value ) = ( $ln =~ /^\s*([^ \t=]+)\s*=\s*(.*\S)\s*$/ ) or die "F:syntax error in config\n";
		$value =~ s/"(.*)"/$1/;#eat quotes if needed
		$settings{"$name"} = $value;
		print "D:reading value of $name:$value\n";
	}

	close CONFIG;

	foreach( ( "darkimage", "lightimage", "brabcnt", "brab_prefix", "brab_suffix", "brab_img_cnt", "brab_anim_time", "workroot", "brab_speed" ) ) {#This must be present
		defined( $settings{$_} ) or die "F:Config value missing for $_\n";
	}

	print "Q[]:cd " . $settings{workroot} . "\n";
	die "F:unable to CD\n" unless( <> =~ /R\[\]:cd\s*\n/ );
}

my( %req, %events, @eventQueue );
my $lastId = 0;

sub hookEvent( $$$$ ) {
	my( $hookCmd, $sub, $single, $tag ) = @_;
	my $id = $lastId ++;
	$events{$id} = {
		hookCmd => $hookCmd,
		code => $sub,
		single => $single,
		tag => $tag
	};
	print "D:Hooking an event: $id/$hookCmd\n";
	print "H[$id]:$hookCmd\n";
}

sub request( $ ) {
	my( $reqCmd ) = @_;
	my( $id ) = $lastId ++;
	print "D:request $id/$reqCmd\n";
	print "Q[$id]:$reqCmd\n";
	until( readInp() eq "req" ) {};#Process it until you find the response to the request
	die( "F:asyncronous response? $req{line}\n" ) if( $id ne $req{id} );
	return( $req{line} );
}

my( $width, $height ) = ( request( "imgsize $settings{lightimage}" ) =~ /imgsize\s+(\d+)\s+(\d+)/ );
my( $antw, $anth ) = ( request( "imgsize $settings{brab_prefix}1$settings{brab_suffix}" ) =~ /imgsize\s+(\d+)\s+(\d+)/ );

print "M:def light\n" .
	"M:placeImg light $settings{lightimage} light -" . ( $width / 2 ) . " -" . ( $height / 2 ) . " 0\n" .
	"M:def dark\n" .
	"M:placeImg dark $settings{darkimage} dark -" . ( $width / 2 ) . " -" . ( $height / 2 ) . " 0\n" .
	"M:def ant\n" .
	"M:placeImg ant $settings{brab_prefix}1$settings{brab_suffix} ant -" . ( $antw / 2 ) . " -" . ( $anth / 2 ) . " 0 globant\n";

for( my $i = 0; $i < 4; $i ++ ) {
	print "M:def ant-$i\n" .
		"M:placeObj ant-$i ant ant 0 0 0\n";
}

print "M:flipX ant-1\n" .
	"M:flipX ant-2\n" .
	"M:flipY ant-2\n" .
	"M:flipY ant-3\n";

for( my $i = 2; $i < $settings{brab_img_cnt}; $i ++ ) {
	print "M:addImg globant $settings{brab_prefix}$i$settings{brab_suffix} $settings{brab_anim_time}\n";
}

my( $wcount, $hcount ) = ( int( 2 * $pixwidth / $width ) + 2, int( 2 * $pixheight / $height ) + 2 );

my @field;
my @fnames = ( "light", "dark" );

sub getCoords( $$ ) {
	my( $x, $y ) = @_;
	$x *= $width;
	$x += $width / 2 if( $y % 2 );
	$y *= $height / 2;
	return( $x, $y );
}

#Setup the field
for( my $x = 0; $x < $wcount; $x ++ ) {
	my @column;
	for( my $y = 0; $y < $hcount; $y ++ ) {
		push @column, my $act = int( rand( 2 ) );#Choose the color
		my( $xcoord, $ycoord ) = getCoords( $x, $y );
		print "M:placeObj root $fnames[$act] f-$x-$y $xcoord $ycoord 100\n";
	}
	push @field, \@column;
}

#An ant looks like this:
#{
#	x => 0,
#	y => 0,
#	dir => 0
#}

my @ants;
my @xdirs = ( -1, 1, 1, -1 );
my @xcmp = ( "<=", ">=", ">=", "<=" );
my @ydirs = ( -1, -1, 1, 1 );
my @ycmp = ( "<=", "<=", ">=", ">=" );
my @turns = ( 1, -1 );

sub stopAnt( $ ) {
	my( $index ) = @_;
	print "M:unplace root ant:$index\n";
}

sub startAnt( $ );

sub antPos( $$ ) {
	my $index = $_[ 1 ];
	stopAnt( $index );
	my $ant = $ants[ $index ];
	my $xuse = ( -1, 1 )[ $ant->{y} % 2 ];
	$ant->{x} += $xdirs[ $ant->{dir} ] if( $xdirs[ $ant->{dir} ] == $xuse );
	$ant->{x} = 0 if( $ant->{x} == $wcount );
	$ant->{x} = $wcount - 1 if( $ant->{x} == -1 );
	$ant->{y} += $ydirs[ $ant->{dir} ];
	$ant->{y} = 0 if( $ant->{y} == $hcount );
	$ant->{y} = $hcount - 1 if( $ant->{y} == -1 );
	$ant->{dir} = ( $ant->{dir} + 4 + $turns[ $field[ $ant->{x} ]->[ $ant->{y} ] ] ) % 4;
	my $act = $field[ $ant->{x} ]->[ $ant->{y} ] = ( 1 + $field[ $ant->{x} ]->[ $ant->{y} ] ) % 2;
	my( $xc, $yc ) = getCoords( $ant->{x}, $ant->{y} );
	print "M:unplace root f-$ant->{x}-$ant->{y}\n" .
		"M:placeObj root $fnames[$act] f-$ant->{x}-$ant->{y} $xc $yc 100\n";
	startAnt( $index );
}

sub startAnt( $ ) {
	my( $index ) = @_;
	my $ant = $ants[ $index ];
	my( $xc, $yc ) = getCoords( $ant->{x}, $ant->{y} );
	print "M:placeObj root ant-" . $ant->{dir} . " ant:$index $xc $yc 0\n" .
		"M:position root ant:$index " . ( $settings{brab_speed} * $xdirs[ $ant->{dir} ] ) . " " . ( $settings{brab_speed} * $ydirs[ $ant->{dir} ] ) . " 1\n";
	hookEvent( "position root ant:$index " . ( $xc + $xdirs[ $ant->{dir} ] * $width / 2 ) .
		" " . ( $yc + $ydirs[ $ant->{dir} ] * $height / 2 ) . " " .
		$xcmp[ $ant->{dir} ] . " " . $ycmp[ $ant->{dir} ], \&antPos, 1, $index );
}

#Setup the ants
for( my $i = 0; $i < $settings{brabcnt}; $i ++ ) {
	sleep 1;
	my %ant = (
		x => int( rand( $wcount ) ),
		y => int( rand( $hcount ) ),
		dir => int( rand( 4 ) )
	);
	push @ants, \%ant;
	startAnt( $i );
}

sub readInp() {
	$_ = <STDIN>;
	return( undef ) unless( defined( $_ ) );
	return( "none" ) if( /^\s*$/ );
	my( $id, $line );
	if( ( $id, $line ) = /E\[([^\[\]]+)\]:(.*)$/ ) {#An event
		if( defined( $events{$id} ) ) {#Does such hook exist?
			my %event = %{$events{$id}};
			$event{line} = $line;
			push( @eventQueue, \%event );#Queue the event for execution
			delete( $events{$id} ) if( $event{single} );#Was it one-happen event?
		} else {
			die( "F:unknown event $id\n" );
		}
		return( "event" );
	} elsif( ( $id, $line ) = /R\[([^\[\]]+)\]:(.*)$/ ) {#A request
		$req{id} = $id;
		$req{line} = $line;
		return( "req" );
	} else {
		die( "F:Unknown command $_\n" );
	}
}

sub processEvents() {
	while( @eventQueue ) {
		my %event = %{shift( @eventQueue )};
		&{$event{code}}( $event{line}, $event{tag} );
	}
}

while( defined( readInp() ) ) {#Process lines until the end of the times
	processEvents();
}
